<?php

/**
* Copyright Maarch since 2022 under license.
* See LICENSE.txt file at the root folder for more details.
* This file is part of Maarch software.
*
*/

/**
* @brief MetaSign Signature Controller
* @author dev@maarch.org
*/

namespace Document\controllers;

use Docserver\controllers\DocserverController;
use Docserver\models\AdrModel;
use Docserver\models\DocserverModel;
use Document\models\DocumentModel;
use SrcCore\models\CoreConfigModel;
use SrcCore\models\CurlModel;
use User\models\UserModel;
use Workflow\models\WorkflowModel;
use SrcCore\models\ValidatorModel;

class MetaSignSignatureController 
{
    public static $signatureLength = 30000;


    /**
     * @description Get config for Metasign signature
     * @return      false|array
     */
    private static function getConfig()
    {
        $loadedXml = CoreConfigModel::getConfig();
        if ($loadedXml->metaSignSignature->enable == 'true') {
            $config = [
                'url'       => (string)$loadedXml->metaSignSignature->url,
                'port'      => (string)$loadedXml->metaSignSignature->port,
                'instance'  => (string)$loadedXml->metaSignSignature->instance,
                'certPath'  => (string)$loadedXml->metaSignSignature->certPath,
                'certPass'  => (string)$loadedXml->metaSignSignature->certPass,
                'certType'  => (string)$loadedXml->metaSignSignature->certType,
                'metaSign'  => [
                    'inDelegationOf' => (string)$loadedXml->metaSignSignature->metaSign->inDelegationOf,
                    'secret' => [
                        "password"   => (string)$loadedXml->metaSignSignature->metaSign->secret->password,
                        "type"       => (string)$loadedXml->metaSignSignature->metaSign->secret->type
                    ],
                    'signatureKeyId'    => (string)$loadedXml->metaSignSignature->metaSign->signatureKeyId,
                    'signatureProfile'  => [
                        'profile' => [
                            'signaturePolicyOid' => (string)$loadedXml->metaSignSignature->metaSign->signatureProfile->signaturePolicyOid,
                            'attachment'         => (string)$loadedXml->metaSignSignature->metaSign->signatureProfile->attachment,
                            'format'             => (string)$loadedXml->metaSignSignature->metaSign->signatureProfile->format,
                            'augmentationLevel'  => (string)$loadedXml->metaSignSignature->metaSign->signatureProfile->augmentationLevel,
                            'signatureAlgoId'    => (string)$loadedXml->metaSignSignature->metaSign->signatureProfile->signatureAlgoId,
                            'archive'            => (string)$loadedXml->metaSignSignature->metaSign->signatureProfile->archive,
                            'requireSigningTime' => (string)$loadedXml->metaSignSignature->metaSign->signatureProfile->requireSigningTime,
                        ]
                    ]
                ]
            ];
            return $config;
        } else {
            return false;
        }
    }

    /**
     * @description Check if the user/certificate can sign documents
     * @return      array|true  errors|true
     */
    public static function init() 
    {
        $config = MetaSignSignatureController::getConfig();
        if(empty($config)) {
            return ['code' => 400, 'errors' => "MetaSgin configuration not found!"];
        }

        $curlResponse = CurlModel::exec([
            'url'       => $config['url'] . (empty($config['port']) ? '' : ":{$config['port']}") . "/servlets/{$config['instance']}/com.bull.security.signserver.rest/signserver/info",
            'method'    => 'GET',
            'headers'   => ['content-type:application/json', 'Accept: application/json'],
            'options'   => [
                CURLOPT_SSLCERT         => $config['certPath'],
                CURLOPT_SSLCERTPASSWD   => $config['certPass'],
                CURLOPT_SSLCERTTYPE     => $config['certType']
            ]
        ]);

        if ($curlResponse['code'] != 200) {
            return ['code' => $curlResponse['code'], 'errors' => $curlResponse['errors']];
        } elseif (empty($curlResponse['response']['methods'])) {
            return ['code' => 500, 'errors' => "MetaSign api endpoints not found"];
        }elseif (!empty($curlResponse['response']['methods']) && !in_array("sigOps/signDoc", $curlResponse['response']['methods'])) {
            return ['code' => 405, 'errors' => "User does not have access to API endpoint to 'sign document' on MetaSign"];
        }
        return true;
    }

    /**
     * @description Get document path from docserver
     * @param   array           $args
     * @return  string|array    filePath|errors
     */
    public static function getDocumentPath(array $args)
    {
        ValidatorModel::notEmpty($args, ['documentId']);
        ValidatorModel::intType($args, ['documentId']);

        $adr = AdrModel::getDocumentsAdr([
            'select' => ['type', 'CONCAT(path, filename) as path_and_filename'],
            'where'  => ['main_document_id = ?', "type IN ('DOC', 'ESIGN')"],
            'data'   => [$args['documentId']]
        ]);
        if (empty($adr)) {
            return ['code' => 404, 'errors' => '[MetaSignSignatureController][getDocumentPath] Document does not exist'];
        }

        $docAdr = [];
        $docTypes = array_column($adr, 'path_and_filename', 'type');

        if (!empty($docTypes['DOC'])) {
            $docAdr = ['type' => 'DOC', 'pathAndFilename' => $docTypes['DOC']];
        }
        if (!empty($docTypes['ESIGN'])) {
            $docAdr = ['type' => 'ESIGN', 'pathAndFilename' => $docTypes['ESIGN']];
        }

        $docserver = DocserverModel::getByType(['type' => $docAdr['type'], 'select' => ['path']]);
        if (empty($docserver['path']) || !file_exists($docserver['path'])) {
            return ['code' => 404, 'errors' => 'Docserver does not exist'];
        }

        $pathToDocument = $docserver['path'] . $docAdr['pathAndFilename'];
        if (!is_file($pathToDocument) || !is_readable($pathToDocument)) {
            return ['code' => 404, 'errors' => 'Document not found on docserver or not readable'];
        }

        return $pathToDocument;
    }

    /**
     * @description Prepare, sign and save document
     * @param   array       $args
     * @return  true|array
     */
    public static function signDocument(array $args)
    {
        ValidatorModel::notEmpty($args, ['documentId']);
        ValidatorModel::intType($args, ['documentId']);
        ValidatorModel::arrayType($args, ['signatureInfo']);

        $config = MetaSignSignatureController::getConfig();
        if(empty($config)) {
            return ['code' => 400, 'errors' => "MetaSign configuration not found!"];
        }

        $pathToDocument = MetaSignSignatureController::getDocumentPath(['documentId' => $args['documentId']]);
        if (!empty($pathToDocument['errors'])) {
            return $pathToDocument;
        }

        $libDir = CoreConfigModel::getLibrariesDirectory();
        if (empty($libDir) || !is_file($libDir . 'SetaPDF-Signer/library/SetaPDF/Autoload.php')) {
            return ['code' => 404, 'errors' => 'SetaPDF-Signer library is not installed'];
        }
        require_once($libDir . 'SetaPDF-Signer/library/SetaPDF/Autoload.php');

        $tmpPath            = CoreConfigModel::getTmpPath();
        $signedDocumentPath = $tmpPath . $GLOBALS['id'] . '_' . rand() . '_signedDocument.pdf';
        $writer             = new \SetaPDF_Core_Writer_File($signedDocumentPath);
        $document           = \SetaPDF_Core_Document::loadByFilename($pathToDocument, $writer);

        $signer = new \SetaPDF_Signer($document);

        $user = UserModel::getById(['id' => $GLOBALS['id'], 'select' => ['firstname', 'lastname', 'email']]);
        $signer->setName($user['firstname'] . ' ' . $user['lastname'] . ' ('. $user['email'] . ')');
        $signer->setReason('Signature du document par ' . $user['firstname'] . ' ' . $user['lastname']);
        $signer->setLocation('Maarch Parapheur');

        $signer->setSignatureContentLength(self::$signatureLength);
        $signer->setAllowSignatureContentLengthChange(true);

        if (!empty($args['signatureInfo'])) {
            $image = \SetaPDF_Core_Image::getByPath($args['signatureInfo']['filePath']);
            $imageXObject = $image->toXObject($document);
            $width = $args['signatureInfo']['signWidth'];
            $height = $imageXObject->getHeight($width);
            $signatureFieldWith = $width > 200 ? $width : 200;

            $fieldName = $signer->addSignatureField(
                'Signature_' . rand(),
                $args['signatureInfo']['page'],
                \SetaPDF_Signer_SignatureField::POSITION_LEFT_TOP,
                ['x' => $args['signatureInfo']['positionX'], 'y' => -$args['signatureInfo']['positionY']],
                $signatureFieldWith,
                $height + 50
            )->getQualifiedName();

            $signer->setSignatureFieldName($fieldName);

            $xObject = \SetaPDF_Core_XObject_Form::create($document, [0, 0, $signatureFieldWith, $height + 50]);
            $canvas = $xObject->getCanvas();
            $imageXObject->draw($canvas, 0, 50, $width, $height);

            $font = new \SetaPDF_Core_Font_Type0_Subset(
                $document,
                __DIR__ . '/fonts/dejavu-fonts-ttf-2.37/ttf/DejaVuSans.ttf'
            );

            $loadedXml = CoreConfigModel::getConfig();
            if ($loadedXml->textWithDigitalSignature == 'true') {
                $textBlock = new \SetaPDF_Core_Text_Block($font, 7);
                $textBlock->setWidth($signatureFieldWith);
                $textBlock->setLineHeight(14);
                $textBlock->setPadding(2);
                $textBlock->setText("Signé électroniquement par : " . $user['firstname'] . ' ' . $user['lastname'] . "\nLe " . date('c'));
                $textBlock->draw($canvas, 0, 30);
            }

            $appearance = new \SetaPDF_Signer_Signature_Appearance_XObject($xObject);
            $signer->setAppearance($appearance);
        } else {
            $fieldName = $signer->addSignatureField()->getQualifiedName();
            $signer->setSignatureFieldName($fieldName);
        }

        $tmpDocument = $signer->preSign(new \SetaPDF_Core_Writer_TempFile());
        
        $fileHash = hash_file('sha256', $tmpDocument->getHashFile()->getPath());
        $binaryHash = hex2bin($fileHash);
        $resultSignature = MetaSignSignatureController::signHash([
            'config'    => $config,
            'docHash'   =>  base64_encode($binaryHash)]);
        if (!empty($resultSignature['errors']) || !empty($resultSignature['code'])) {
            return ['code' => $resultSignature['code'], 'errors' => $resultSignature['errors']];
        }

        try {
            $signer->saveSignature($tmpDocument, $resultSignature['signatureContent']);
        } catch (\SetaPDF_Signer_Exception_ContentLength $e) {
            self::$signatureLength = self::$signatureLength + 1000;
            MetaSignSignatureController::signDocument($args);
            return true;
        }

        $storeResInfos = DocserverController::storeResourceOnDocServer([
            'encodedFile'   => base64_encode(file_get_contents($signedDocumentPath)),
            'format'        => 'pdf',
            'docserverType' => 'ESIGN'
        ]);

        if (!empty($storeResInfos['errors'])) {
            return ['errors' => $storeResInfos['errors']];
        }

        AdrModel::deleteDocumentAdr([
            'where' => ['main_document_id = ?', 'type = ?'],
            'data'  => [$args['documentId'], 'ESIGN']
        ]);
        AdrModel::createDocumentAdr([
            'documentId'  => $args['documentId'],
            'type'        => 'ESIGN',
            'path'        => $storeResInfos['path'],
            'filename'    => $storeResInfos['filename'],
            'fingerprint' => $storeResInfos['fingerprint']
        ]);
        unlink($signedDocumentPath);

        return true;
    }

    /**
     * @description Send document hash to be signed from MetaSign server
     * @param   array   $args
     * @return  array   Return signature content from MetaSign server
     */
    private static function signHash(array $args)
    {
        ValidatorModel::notEmpty($args, ['docHash', 'config']);
        ValidatorModel::stringType($args, ['docHash']);
        ValidatorModel::arrayType($args, ['config']);

        $curlResponse = CurlModel::exec([
            'url'       => $args['config']['url'] . (empty($args['config']['port']) ? '' : ":{$args['config']['port']}") . "/servlets/{$args['config']['instance']}/com.bull.security.signserver.rest/signserver/sigOps/signDoc",
            'method'    => 'POST',
            'headers'   => ['content-type:application/json', 'Accept: application/json'],
            'options'   => [
                CURLOPT_SSLCERT         => $args['config']['certPath'],
                CURLOPT_SSLCERTPASSWD   => $args['config']['certPass'],
                CURLOPT_SSLCERTTYPE     => $args['config']['certType']
            ],
            'body'      => json_encode([
                "inDelegationOf" => $args['config']['metaSign']['inDelegationOf'],
                "secret" => [
                    "password" => $args['config']['metaSign']['secret']['password'],
                    "type"  => $args['config']['metaSign']['secret']['type']
                ],
                "signatureKeyId" => $args['config']['metaSign']['signatureKeyId'],
                "signatureProfile" => $args['config']['metaSign']['signatureProfile'],
                "document" => [
                    "docHash"  => $args['docHash']
                ]
            ])
        ]);

        if ($curlResponse['code'] != 200) {
            return ['code' => $curlResponse['code'], 'errors' => $curlResponse['response']['errorInfo'] ?? $curlResponse['errors']];
        }

        return ['signatureContent' => base64_decode($curlResponse['response']['signatureContent']['bytes'])];
    }
}