<?php

/**
* Copyright Maarch since 2008 under license.
* See LICENSE.txt file at the root folder for more details.
* This file is part of Maarch software.
*
*/

/**
* @brief Notifications Schedule Controller
* @author dev@maarch.org
*/

namespace Notification\controllers;

use Configuration\models\ConfigurationModel;
use Notification\models\NotificationsScheduleModel;
use Group\controllers\PrivilegeController;
use History\controllers\HistoryController;
use Respect\Validation\Validator;
use Slim\Psr7\Request;
use SrcCore\http\Response;
use SrcCore\models\CoreConfigModel;
use User\models\UserModel;

class NotificationsScheduleController
{
    public const TYPES = ['next_user', 'typist_END', 'typist_INT', 'typist_DEL', 'typist_REF'];
    public const FREQUENCY_MODES = ['minutes', 'hours'];

    public function create(Request $request, Response $response)
    {
        if (!PrivilegeController::hasPrivilege(['userId' => $GLOBALS['id'], 'privilege' => 'manage_notifications'])) {
            return $response->withStatus(403)->withJson(['errors' => 'Privilege forbidden']);
        }

        $body = $request->getParsedBody();
        // label, type, months, daysOfMonth, daysOfWeek, startTime, (endTime, frequency, frequencyMode)?

        $err = NotificationsScheduleController::validateBody($body);
        if (!empty($err)) {
            return $response->withStatus(400)->withJson(['errors' => $err]);
        }

        $body['months'] = array_values(array_unique($body['months']));
        $body['daysOfMonth'] = array_values(array_unique($body['daysOfMonth']));
        $body['daysOfWeek'] = array_values(array_unique($body['daysOfWeek']));
        sort($body['months']);
        sort($body['daysOfMonth']);
        sort($body['daysOfWeek']);

        $id = NotificationsScheduleModel::create([
            'label'          => $body['label'],
            'type'           => $body['type'],
            'months'         => empty($body['months']) ? '[]' : json_encode($body['months']),
            'days_of_month'  => empty($body['daysOfMonth']) ? '[]' : json_encode($body['daysOfMonth']),
            'days_of_week'   => empty($body['daysOfWeek']) ? '[]' : json_encode($body['daysOfWeek']),
            'start_time'     => $body['startTime'],
            'end_time'       => $body['endTime'],
            'frequency'      => $body['frequency'],
            'frequency_mode' => $body['frequencyMode'],
        ]);

        HistoryController::add([
            'code'          => 'OK',
            'objectType'    => 'notifications_schedule',
            'objectId'      => $id,
            'type'          => 'CREATION',
            'message'       => "{notificationScheduleItemAdded} : {$body['label']}, type {$body['type']}"
        ]);

        return $response->withJson(['id' => $id]);
    }

    public function get(Request $request, Response $response)
    {
        $rawNotificationsSchedule = NotificationsScheduleModel::get();

        $notificationsSchedule = [];
        foreach ($rawNotificationsSchedule as $value) {
            $notificationsSchedule[] = [
                'id'            => $value['id'],
                'label'         => $value['label'],
                'type'          => $value['type'],
                'months'        => json_decode($value['months']),
                'daysOfMonth'   => json_decode($value['days_of_month']),
                'daysOfWeek'    => json_decode($value['days_of_week']),
                'startTime'     => (new \DateTime($value['start_time'] ?? ''))->format('H:i'),
                'endTime'       => (new \DateTime($value['end_time'] ?? ''))->format('H:i'),
                'frequency'     => $value['frequency'],
                'frequencyMode' => $value['frequency_mode'],
                'status'        => $value['status'],
                'lastExecution' => $value['last_execution'],
            ];
        }

        return $response->withJson(['notificationsSchedule' => $notificationsSchedule]);
    }

    public function getById(Request $request, Response $response, array $args)
    {
        if (!PrivilegeController::hasPrivilege(['userId' => $GLOBALS['id'], 'privilege' => 'manage_notifications'])) {
            return $response->withStatus(403)->withJson(['errors' => 'Privilege forbidden']);
        }
        if (!Validator::intVal()->notEmpty()->validate($args['id'])) {
            return $response->withStatus(400)->withJson(['errors' => 'Route id must be an integer']);
        }

        $rawNotificationsScheduleItem = NotificationsScheduleModel::getById(['id' => $args['id']]);
        if (empty($rawNotificationsScheduleItem)) {
            return $response->withStatus(400)->withJson(['errors' => 'Notifications schedule item not found']);
        }

        $notificationsScheduleItem = [
            'id'            => $rawNotificationsScheduleItem['id'],
            'label'         => $rawNotificationsScheduleItem['label'],
            'type'          => $rawNotificationsScheduleItem['type'],
            'months'        => json_decode($rawNotificationsScheduleItem['months']),
            'daysOfMonth'   => json_decode($rawNotificationsScheduleItem['days_of_month']),
            'daysOfWeek'    => json_decode($rawNotificationsScheduleItem['days_of_week']),
            'startTime'     => (new \DateTime($rawNotificationsScheduleItem['start_time']))->format('H:i'),
            'endTime'       => !empty($rawNotificationsScheduleItem['end_time']) ? (new \DateTime($rawNotificationsScheduleItem['end_time']))->format('H:i') : null,
            'frequency'     => $rawNotificationsScheduleItem['frequency'],
            'frequencyMode' => $rawNotificationsScheduleItem['frequency_mode'],
            'status'        => $rawNotificationsScheduleItem['status'],
            'lastExecution' => $rawNotificationsScheduleItem['last_execution'],
        ];

        return $response->withJson(['notificationsScheduleItem' => $notificationsScheduleItem]);
    }

    public function update(Request $request, Response $response, array $args)
    {
        if (!PrivilegeController::hasPrivilege(['userId' => $GLOBALS['id'], 'privilege' => 'manage_notifications'])) {
            return $response->withStatus(403)->withJson(['errors' => 'Privilege forbidden']);
        }

        if (!Validator::intVal()->notEmpty()->validate($args['id'])) {
            return $response->withStatus(400)->withJson(['errors' => 'Route id must be an integer']);
        }

        $body = $request->getParsedBody();

        $err = NotificationsScheduleController::validateBody($body);
        if (!empty($err)) {
            return $response->withStatus(400)->withJson(['errors' => $err]);
        }

        $body['months'] = array_values(array_unique($body['months']));
        $body['daysOfMonth'] = array_values(array_unique($body['daysOfMonth']));
        $body['daysOfWeek'] = array_values(array_unique($body['daysOfWeek']));
        sort($body['months']);
        sort($body['daysOfMonth']);
        sort($body['daysOfWeek']);

        $notificationsScheduleItem = NotificationsScheduleModel::getById(['id' => $args['id']]);
        if (empty($notificationsScheduleItem)) {
            return $response->withStatus(400)->withJson(['errors' => 'Notifications schedule item not found']);
        }

        NotificationsScheduleModel::update([
            'set' => [
                'label'          => $body['label'],
                'type'           => $body['type'],
                'months'         => empty($body['months']) ? '[]' : json_encode($body['months']),
                'days_of_month'  => empty($body['daysOfMonth']) ? '[]' : json_encode($body['daysOfMonth']),
                'days_of_week'   => empty($body['daysOfWeek']) ? '[]' : json_encode($body['daysOfWeek']),
                'start_time'     => $body['startTime'],
                'end_time'       => $body['endTime'],
                'frequency'      => $body['frequency'],
                'frequency_mode' => $body['frequencyMode'],
                'status'         => $body['status'],
                'last_execution' => $body['lastExecution'],
            ],
            'where' => ['id = ?'],
            'data'  => [$args['id']]
        ]);

        HistoryController::add([
            'code'       => 'OK',
            'objectType' => 'notifications_schedule',
            'objectId'   => $args['id'],
            'type'       => 'MODIFICATION',
            'message'    => "{notificationsScheduleItemUpdated} : {$body['label']}, type {$body['type']}"
        ]);

        return $response->withStatus(204);
    }

    public function delete(Request $request, Response $response, array $args)
    {
        if (!PrivilegeController::hasPrivilege(['userId' => $GLOBALS['id'], 'privilege' => 'manage_notifications'])) {
            return $response->withStatus(403)->withJson(['errors' => 'Privilege forbidden']);
        }

        if (!Validator::intVal()->notEmpty()->validate($args['id'])) {
            return $response->withStatus(400)->withJson(['errors' => 'Id must be an integer']);
        }

        $notificationsScheduleItem = NotificationsScheduleModel::getById(['id' => $args['id']]);
        if (empty($notificationsScheduleItem)) {
            return $response->withStatus(400)->withJson(['errors' => 'Notifications schedule item not found']);
        }

        NotificationsScheduleModel::delete(['where' => ['id = ?'], 'data' => [$args['id']]]);

        $users = UserModel::get([
            'select' => ['id', 'preferences'],
            'where'  => ["preferences->'notifications'->'summaries' != '[]'"]
        ]);

        if (!empty($users)) {
            foreach ($users as $user) {
                $user['preferences'] = json_decode($user['preferences'], true);
                if (in_array($args['id'], $user['preferences']['notifications']['summaries'])) {
                    $scheduleIndex = array_search($args['id'], $user['preferences']['notifications']['summaries']);
                    array_splice($user['preferences']['notifications']['summaries'], $scheduleIndex, 1);
                    $user['preferences'] = json_encode($user['preferences']);
                    UserModel::update(['set' => ['preferences' => $user['preferences']], 'where' => ['id = ?'], 'data' => [$user['id']]]);
                }
            }
        }

        HistoryController::add([
            'code'          => 'OK',
            'objectType'    => 'notifications_schedule',
            'objectId'      => $args['id'],
            'type'          => 'SUPPRESSION',
            'message'       => "{notificationsScheduleItemDeleted} : {$notificationsScheduleItem['label']}, type {$notificationsScheduleItem['type']}"
        ]);

        return $response->withStatus(204);
    }

    public function initializeSchedulerScript(Request $request, Response $response)
    {
        if (!PrivilegeController::hasPrivilege(['userId' => $GLOBALS['id'], 'privilege' => 'manage_notifications'])) {
            return $response->withStatus(403)->withJson(['errors' => 'Privilege forbidden']);
        }

        $configuration = ConfigurationModel::getByIdentifier(['identifier' => 'customization']);
        $configuration = $configuration[0] ?? null;
        $configuration['value'] = json_decode($configuration['value'] ?? '{}', true);

        if (empty($configuration['value']['schedulerPath']) || !is_file($configuration['value']['schedulerPath'])) {
            $schedulerScript = NotificationsScheduleController::createSchedulerScript();
            if (!empty($schedulerScript['errors'])) {
                return $response->withStatus(500)->withJson(['errors' => $schedulerScript['errors']]);
            }
            $configuration['value']['schedulerPath'] = $schedulerScript['path'];
        }

        // Check if script already in crontab
        exec('crontab -l|grep ' . $configuration['value']['schedulerPath'], $output);
        if (empty($output)) {
            // run scheduler every minutes
            $crontabLine = '*/1 * * * * ' . $configuration['value']['schedulerPath'];

            exec('(crontab -l; echo "' . $crontabLine . '")|crontab -');

            HistoryController::add([
                'code'       => 'OK',
                'objectType' => 'configurations',
                'objectId'   => $configuration['id'] ?? null,
                'type'       => 'MODIFICATION',
                'message'    => "{schedulerScriptCreated} : {$configuration['value']['schedulerPath']}"
            ]);
        }

        return $response->withStatus(204);
    }

    public function checkSchedulerScriptInitialized(Request $request, Response $response)
    {
        if (!PrivilegeController::hasPrivilege(['userId' => $GLOBALS['id'], 'privilege' => 'manage_notifications'])) {
            return $response->withStatus(403)->withJson(['errors' => 'Privilege forbidden']);
        }

        $configuration = ConfigurationModel::getByIdentifier(['identifier' => 'customization']);
        $configuration = $configuration[0] ?? null;
        $configuration = json_decode($configuration['value'] ?? '{}', true);

        if (empty($configuration['schedulerPath']) || !is_file($configuration['schedulerPath'])) {
            return $response->withJson(['schedulerInitialized' => false]);
        }

        exec('crontab -l|grep ' . $configuration['schedulerPath'], $output);

        if (empty($output)) {
            return $response->withJson(['schedulerInitialized' => false]);
        }

        return $response->withJson(['schedulerInitialized' => true]);
    }

    public function deleteSchedulerScript(Request $request, Response $response)
    {
        if (!PrivilegeController::hasPrivilege(['userId' => $GLOBALS['id'], 'privilege' => 'manage_notifications'])) {
            return $response->withStatus(403)->withJson(['errors' => 'Privilege forbidden']);
        }

        $configuration = ConfigurationModel::getByIdentifier(['identifier' => 'customization']);
        $configuration = $configuration[0];
        $configuration['value'] = json_decode($configuration['value'], true);

        if (empty($configuration['value']['schedulerPath'])) {
            return $response->withStatus(204);
        }

        if (is_file($configuration['value']['schedulerPath'])) {
            unlink($configuration['value']['schedulerPath']);
        }

        exec('crontab -l|grep ' . $configuration['value']['schedulerPath'], $output);
        if (empty($output)) {
            return $response->withStatus(204);
        }

        exec('(crontab -l| grep -v "' . $configuration['value']['schedulerPath'] . '")|crontab -');

        $configuration['value']['schedulerPath'] = '';
        $configuration['value'] = json_encode($configuration['value']);
        ConfigurationModel::update(['set' => ['value' => $configuration['value']], 'where' => ['identifier = ?'], 'data' => ['customization']]);

        HistoryController::add([
            'code'       => 'OK',
            'objectType' => 'configurations',
            'objectId'   => $configuration['id'],
            'type'       => 'SUPPRESSION',
            'message'    => "{schedulerScriptDeleted} : {$configuration['value']['schedulerPath']}"
        ]);

        return $response->withStatus(204);
    }

    private static function createSchedulerScript()
    {
        $filename = 'scheduler_' . rand() . '.sh';

        $corePath = str_replace('src/app/notifications/controllers', '', __DIR__);
        $corePath = str_replace('src/app/notification/models', '', $corePath);

        $configPath = CoreConfigModel::getConfigPath();

        $fullPath = $corePath . 'bin/' . $filename;

        if (!is_writable($corePath . 'bin/')) {
            return ['errors' => 'Cannot create scheduler script in ' . $corePath . 'bin/'];
        }

        $file_open = fopen($fullPath, 'w+');

        fwrite($file_open, '#!/bin/sh');
        fwrite($file_open, "\n");
        fwrite($file_open, 'path=\''.$corePath.'\'');
        fwrite($file_open, "\n");
        fwrite($file_open, 'cd $path');
        fwrite($file_open, "\n");
        fwrite($file_open, 'filePath=\''.$corePath.'bin/scheduler.php\'');
        fwrite($file_open, "\n");
        fwrite($file_open, 'php $filePath '.$configPath);
        fwrite($file_open, "\n");
        fclose($file_open);
        shell_exec('chmod +x '.escapeshellarg($fullPath));

        $configuration = ConfigurationModel::getByIdentifier(['identifier' => 'customization']);
        if (empty($configuration)) {
            return ['errors' => 'Configuration not found'];
        }
        $configuration = $configuration[0] ?? null;
        $configuration['value'] = json_decode($configuration['value'] ?? '{}', true);

        $configuration['value']['schedulerPath'] = $fullPath;
        $configuration['value'] = json_encode($configuration['value']);
        ConfigurationModel::update(['set' => ['value' => $configuration['value']], 'where' => ['id = ?'], 'data' => [$configuration['id']]]);

        return ['path' => $fullPath];
    }

    private static function validateBody($body)
    {
        if (empty($body)) {
            return 'Body is not set or empty';
        } elseif (!Validator::stringType()->length(1, 256)->validate($body['label'])) {
            return 'Body label is too short or too long (must be between 1 and 256 characters)';
        } elseif (!Validator::stringType()->in(NotificationsScheduleController::TYPES)->validate($body['type'])) {
            return 'Body type does not exist';
        } elseif (!Validator::arrayType()->oneOf(Validator::equals(['*']), Validator::each(Validator::intVal()->between(1, 12)))->validate($body['months'])) {
            return "Body months must contain numbers from 1-12 or only '*'";
        } elseif (!Validator::arrayType()->oneOf(Validator::equals(['*']), Validator::each(Validator::intVal()->between(1, 31)))->validate($body['daysOfMonth'])) {
            return "Body daysOfMonth must contain numbers from 1-31 or only '*'";
        } elseif (!Validator::arrayType()->oneOf(Validator::equals(['*']), Validator::each(Validator::intVal()->between(1, 7)))->validate($body['daysOfWeek'])) {
            return "Body daysOfWeek must contain numbers from 1-7 or only '*'";
        } elseif (!Validator::equals(['*'])->validate($body['months']) && !Validator::equals(['*'])->validate($body['daysOfWeek'])) {
            return 'At least one of body months and daysOfWeek must be [\'*\']';
        } elseif (!Validator::stringType()->time('H:i')->validate($body['startTime'])) {
            return 'Body startTime is not a valid time, expecting hh:mm';
        }
        if (!empty($body['frequency'])) {
            if (!Validator::stringType()->time('H:i')->greaterThan($body['startTime'])->validate($body['endTime'])) {
                return 'Body endTime is not a valid time or is before startTime, expecting hh:mm';
            } elseif (!Validator::intType()->positive()->validate($body['frequency'])) {
                return 'Body frequency is not a number or is less than 0';
            } elseif (!Validator::stringType()->in(NotificationsScheduleController::FREQUENCY_MODES)->validate($body['frequencyMode'])) {
                return "Body frequencyMode must be 'minutes' or 'hours'";
            }
        }
        return null;
    }
}
