<?php

/**
* Copyright Maarch since 2008 under license.
* See LICENSE.txt file at the root folder for more details.
* This file is part of Maarch software.
*
*/

/**
* @brief FastOTP Controller
* @author dev@maarch.org
*/

namespace Workflow\controllers;

use Workflow\models\WorkflowExternalInformationModel;
use SrcCore\models\CurlModel;
use ExternalSignatoryBook\models\ExternalSignatoryBookModel;
use SrcCore\models\AuthenticationModel;
use Respect\Validation\Validator;
use SrcCore\models\ValidatorModel;
use Configuration\models\ConfigurationModel;
use Slim\Psr7\Request;
use SrcCore\http\Response;
use Workflow\models\WorkflowModel;
use Document\controllers\DocumentController;
use Docserver\controllers\DocserverController;
use Docserver\models\AdrModel;
use SrcCore\models\CoreConfigModel;
use Firebase\JWT\JWT;
use Document\models\DocumentModel;

class FastOTPController
{
    public function getReturnFast(Request $request, Response $response): Response
    {
        $queryParams = $request->getQueryParams();

        $checkToken = FastOTPController::checkToken(['token' => $queryParams['token'] ?? null]);
        if (!empty($checkToken['errors'])) {
            return $response->withStatus(400)->withJson(['errors' => $checkToken['errors']]);
        }
        $documentId = $checkToken['documentId'];
        $workflowId = $checkToken['workflowId'];
        $fastStatus = $queryParams['status'] ?? null;

        $document = DocumentModel::getById(['select' => ['id', 'title', 'reference'], 'id' => $documentId]);
        if (empty($document)) {
            return $response->withStatus(400)->withJson(['errors' => 'Document does not exist']);
        }

        $workflow = WorkflowModel::get([
            'select' => ['status'],
            'where'  => ['main_document_id = ?', 'id = ?'],
            'data'   => [$documentId, $workflowId]
        ]);
        if (empty($workflow[0])) {
            return $response->withStatus(403)->withJson(['errors' => 'Workflow does not exist']);
        } elseif (!empty($workflow[0]['status']) && $workflow[0]['status'] == 'STOP') {
            return $response->withStatus(403)->withJson(['errors' => 'Workflow is interrupted', 'lang' => 'workflowInterrupted']);
        }

        $workflowExternalInformations = WorkflowExternalInformationModel::getByWorkflowId(['select' => ['informations', 'external_signatory_book_id'], 'workflowId' => $workflowId]);
        if (empty($workflowExternalInformations)) {
            return $response->withStatus(403)->withJson(['errors' => 'External Workflow info does not exist']);
        }

        $informations = json_decode($workflowExternalInformations['informations'], true);
        if (empty($informations['fastDocId'])) {
            return $response->withStatus(403)->withJson(['errors' => 'Fast document id not saved']);
        }

        $externalSB = ExternalSignatoryBookModel::getById(['select' => ['*'], 'id' => $workflowExternalInformations['external_signatory_book_id']]);
        if (empty($externalSB)) {
            return $response->withStatus(400)->withJson(['errors' => 'No external signatory book configuration found']);
        } elseif ($externalSB['type'] != 'fast') {
            return $response->withStatus(400)->withJson(['errors' => 'External signatory book configuration is invalid']);
        }
        $externalSB['connection_data'] = json_decode($externalSB['connection_data'], true);

        $comments = null;
        if (!empty($fastStatus) && $fastStatus == 'SIGNED') {

            $fileResponse = FastOTPController::download([
                'connection_data' => $externalSB['connection_data'],
                'documentId' => $informations['fastDocId']
            ]);
            if (!empty($fileResponse['errors'])) {
                return $response->withStatus(400)->withJson(['errors' => [$fileResponse['errors']]]);
            }

            $storeInfos = DocserverController::storeResourceOnDocServer([
                'encodedFile'     => $fileResponse['b64FileContent'],
                'format'          => 'pdf',
                'docserverType'   => 'DOC'
            ]);
            if (!empty($storeInfos['errors'])) {
                return $response->withStatus(500)->withJson(['errors' => $storeInfos['errors']]);
            }

            AdrModel::deleteDocumentAdr([
                'where' => ['main_document_id = ?', 'type = ?'],
                'data'  => [$documentId, 'DOC']
            ]);
            AdrModel::createDocumentAdr([
                'documentId'    => $documentId,
                'type'          => 'DOC',
                'path'          => $storeInfos['path'],
                'filename'      => $storeInfos['filename'],
                'fingerprint'   => $storeInfos['fingerprint']
            ]);

            $storeInfos = DocserverController::storeResourceOnDocServer([
                'encodedFile'     => $fileResponse['b64FileContent'],
                'format'          => 'pdf',
                'docserverType'   => 'ESIGN'
            ]);
            if (!empty($storeInfos['errors'])) {
                return $response->withStatus(500)->withJson(['errors' => $storeInfos['errors']]);
            }

            AdrModel::deleteDocumentAdr([
                'where' => ['main_document_id = ?', 'type = ?'],
                'data'  => [$documentId, 'ESIGN']
            ]);
            AdrModel::createDocumentAdr([
                'documentId'    => $documentId,
                'type'          => 'ESIGN',
                'path'          => $storeInfos['path'],
                'filename'      => $storeInfos['filename'],
                'fingerprint'   => $storeInfos['fingerprint']
            ]);

            AdrModel::deleteDocumentAdr([
                'where' => ['main_document_id = ?', 'type like ?'],
                'data'  => [$documentId, 'TNL%']
            ]);
            $configPath = CoreConfigModel::getConfigPath();
            exec("php src/app/convert/scripts/ThumbnailScript.php '{$configPath}' {$documentId} 'document' 0 > /dev/null &");

        } elseif (!empty($fastStatus) && $fastStatus == 'REFUSED') {

            $messageResponse = FastOTPController::getRefusalMessage([
                'connection_data'   => $externalSB['connection_data'],
                'documentId'        => $informations['fastDocId'],
            ]);
            if (!empty($messageResponse['errors'])) {
                return $response->withStatus(400)->withJson(['errors' => $messageResponse['errors']]);
            }
            $deleteResponse = FastOTPController::delete([
                'connection_data'   => $externalSB['connection_data'],
                'documentId'        => $informations['fastDocId'],
            ]);
            if (!empty($deleteResponse['errors'])) {
                return $response->withStatus(400)->withJson(['errors' => $deleteResponse['errors']]);
            }

            $comments = $messageResponse['response'];

        } elseif (!empty($fastStatus) && $fastStatus == 'ERROR') {

            $informations['errors']['status'] = $fastStatus;
            $informations['errors']['logs']   = "An error on FastParapheur";
            $informations = json_encode($informations);
            WorkflowExternalInformationModel::update(['set' => ['informations' => $informations], 'where' => ['workflow_id = ?'], 'data' => [$workflowId]]);

            return $response->withStatus(400)->withJson(['errors' => "FastParapheur return an error"]);

        } elseif (!empty($fastStatus) && $fastStatus == 'NONE') {

            $informations['errors']['status'] = $fastStatus;
            $informations['errors']['logs']   = "A click on the redirect button without signing or refusing the document";
            $informations = json_encode($informations);
            WorkflowExternalInformationModel::update(['set' => ['informations' => $informations], 'where' => ['workflow_id = ?'], 'data' => [$workflowId]]);

            return $response->withStatus(400)->withJson(['errors' => "FastParapheur return no action taken"]);
        } else {
            return $response->withStatus(400)->withJson(['errors' => "Invalid 'fastStatus'"]);
        }

        $GLOBALS['webhook'] = 'fastParapheur redirection';
        $GLOBALS['id'] = 0;
        DocumentController::endAction([
            'id'            => $documentId,
            'workflowId'    => $workflowId,
            'status'        => $fastStatus == 'SIGNED' ? 'VAL' : 'REF',
            'note'          => $comments ?? null
        ]);

        return $response->withJson([
            'docId'        => $document['id'],
            'docObjet'     => $document['title'],
            'docReference' => $document['reference']
        ]);
    }

    public static function createProcedure(array $args)
    {
        ValidatorModel::notEmpty($args, ['documentId', 'encodedDocument', 'name', 'workflowId']);
        ValidatorModel::stringType($args, ['encodedDocument', 'name', 'description']);
        ValidatorModel::intVal($args, ['documentId', 'workflowId']);

        $workflowExternalInformations = WorkflowExternalInformationModel::getByWorkflowId(['select' => ['*'], 'workflowId' => $args['workflowId']]);
        if (empty($workflowExternalInformations)) {
            return ['errors' => 'No external informations found'];
        }
        $workflowExternalInformations['informations'] = json_decode($workflowExternalInformations['informations'], true);

        $externalSB = ExternalSignatoryBookModel::getById(['select' => ['*'], 'id' => $workflowExternalInformations['external_signatory_book_id']]);
        if (empty($externalSB)) {
            return ['errors' => 'No external signatory book configuration found'];
        } elseif ($externalSB['type'] !== 'fast') {
            return ['errors' => 'External signatory book configuration is invalid'];
        }
        $externalSB['connection_data'] = json_decode($externalSB['connection_data'], true);

        $check = FastOTPController::controlConnectionDataConfig(['connection_data' => $externalSB['connection_data']]);
        if (!empty($check['errors'])) {
            return ['errors' => $check['errors']];
        }

        $otpXml = FastOTPController::generateOtpXml([
            'prettyPrint'   => true,
            'otpInfo'       => [
                'OTP_firstname_0'   => $workflowExternalInformations['firstname'],
                'OTP_lastname_0'    => $workflowExternalInformations['lastname'],
                'OTP_phonenumber_0' => $workflowExternalInformations['phone'],
                'OTP_email_0'       => $workflowExternalInformations['email']
            ]
        ]);
        if (!empty($otpXml['errors'])) {
            return ['errors' => $otpXml['errors']];
        }

        $today = new \DateTimeImmutable();
        $token = FastOTPController::generateToken([
            'payload' => [
                'documentId' => $args['documentId'],
                'workflowId' => $args['workflowId']
            ]
        ]);

        $multipartBody = [
            'document'   => [
                'isFile'   => true,
                'filename' => "{$args['name']}_{$today->format('d-m-Y_H-i-s')}.pdf",
                'content'  => base64_decode($args['encodedDocument'])
            ],
            'meta' => [
                'isFile'   => true,
                'filename' => 'METAS_API.xml',
                'content'  => $otpXml['content']
            ],
            'referrer' => ConfigurationModel::getApplicationUrl() . "dist/#/externalSignatoryBookReturn/token/{$token}"
        ];

        $response = FastOTPController::uploadOTP([
            'connection_data'   => $externalSB['connection_data'],
            'multipartBody'     => $multipartBody
        ]);
        if (!empty($response['errors'])) {
            FastOTPController::updateErrorInfo([
                'workflowId'   => $args['workflowId'],
                'informations' => ['errors' => $response['errors']]
            ]);
            return ['errors' => $response['errors']];
        }

        $workflowExternalInformations['informations']['token'] = $token;
        $workflowExternalInformations['informations']['fastDocId'] = $response['response']['id'];
        $workflowExternalInformations['informations']['fastUrl']   = $response['response']['url'];
        $informations = json_encode($workflowExternalInformations['informations']);

        WorkflowExternalInformationModel::update(['set' => ['informations' => $informations], 'where' => ['workflow_id = ?'], 'data' => [$args['workflowId']]]);

        return true;
    }

    public static function deleteProcedureByWorkflowId(array $args)
    {
        ValidatorModel::notEmpty($args, ['workflowId']);
        ValidatorModel::intVal($args, ['workflowId']);

        $workflowExternalInformations = WorkflowExternalInformationModel::getByWorkflowId(['select' => ['informations', 'external_signatory_book_id'], 'workflowId' => $args['workflowId']]);
        if (empty($workflowExternalInformations)) {
            return ['errors' => 'Workflow does not exist'];
        }

        $informations = json_decode($workflowExternalInformations['informations'], true);

        $externalSB = ExternalSignatoryBookModel::getById(['select' => ['*'], 'id' => $workflowExternalInformations['external_signatory_book_id']]);
        if (empty($externalSB)) {
            return ['errors' => 'No external signatory book configuration found'];
        } elseif ($externalSB['type'] != 'fast') {
            return ['errors' => 'External signatory book configuration is invalid'];
        }
        $externalSB['connection_data'] = json_decode($externalSB['connection_data'], true);


        if (!empty($informations['fastDocId'])) {
            unset($informations['token']);
            unset($informations['fastDocId']);
            unset($informations['fastUrl']);

            $informations = json_encode($informations);
            WorkflowExternalInformationModel::update(['set' => ['informations' => $informations], 'where' => ['workflow_id = ?'], 'data' => [$args['workflowId']]]);
        }

        return true;
    }

    public static function controlConnectionDataConfig(array $args)
    {
        if (!Validator::arrayType()->notEmpty()->validate($args['connection_data'])) {
            return ['errors' => "External signatory book configuration 'connection_data' is empty or not a array"];
        } elseif (!Validator::stringType()->notEmpty()->validate($args['connection_data']['subscriberId'])) {
            return ['errors' => "External signatory book configuration 'subscriberId' is empty or not a string"];
        } elseif (!Validator::stringType()->notEmpty()->validate($args['connection_data']['apiUri'])) {
            return ['errors' => "External signatory book configuration 'apiUri' is empty or not a string"];
        } elseif (!Validator::stringType()->notEmpty()->validate($args['connection_data']['certPath'])) {
            return ['errors' => "External signatory book configuration 'certPath' is empty or not a string"];
        } elseif (!file_exists($args['connection_data']['certPath'])) {
            return ['errors' => "External signatory book configuration 'certPath' : File does not exist"];
        } elseif (!Validator::stringType()->notEmpty()->validate($args['connection_data']['certPass'])) {
            return ['errors' => "External signatory book configuration 'certPass' is empty or not a string"];
        } elseif (!Validator::stringType()->notEmpty()->validate($args['connection_data']['certType'])) {
            return ['errors' => "External signatory book configuration 'certType' is empty or not a string"];
        }

        return true;
    }

    public static function updateErrorInfo(array $args)
    {
        ValidatorModel::notEmpty($args, ['workflowId', 'informations']);
        ValidatorModel::intVal($args, ['workflowId']);
        ValidatorModel::arrayType($args, ['informations']);

        WorkflowExternalInformationModel::update([
            'set' => [
                'informations' => json_encode($args['informations'])
            ],
            'where' => ['workflow_id = ?'],
            'data'  => [$args['workflowId']]
        ]);
        return true;
    }

    public static function uploadOTP(array $args)
    {
        ValidatorModel::notEmpty($args, ['connection_data', 'multipartBody']);
        ValidatorModel::arrayType($args, ['connection_data', 'multipartBody']);

        $curlReturn = CurlModel::exec([
            'method'    => 'POST',
            'url'       => $args['connection_data']['apiUri'] . "/documents/v2/otp/{$args['connection_data']['subscriberId']}/signature-otp/upload",
            'options'   => [
                CURLOPT_SSLCERT         => $args['connection_data']['certPath'],
                CURLOPT_SSLCERTPASSWD   => AuthenticationModel::decrypt(['cryptedPassword' => $args['connection_data']['certPass']]),
                CURLOPT_SSLCERTTYPE     => $args['connection_data']['certType']
            ],
            'multipartBody' => $args['multipartBody']
        ]);

        if ($curlReturn['code'] != 200 || !empty($curlReturn['response']['errorCode'])) {
            if (!empty($curlReturn['response']['userFriendlyMessage']) && $curlReturn['response']['userFriendlyMessage'] != 'Erreur') {
                return ['errors' => $curlReturn['response']['userFriendlyMessage']];
            } else {
                return ['errors' => $curlReturn['response']['developerMessage']];
            }
        }
        return ['response' => $curlReturn['response']];
    }

    public static function generateOtpXml(array $args)
    {
        ValidatorModel::notEmpty($args, ['otpInfo']);
        ValidatorModel::arrayType($args, ['otpInfo']);
        ValidatorModel::boolType($args, ['prettyPrint']);

        $otpInfoXML = new \SimpleXMLElement('<?xml version="1.0" encoding="UTF-8" standalone="yes"?> <meta-data-list></meta-data-list>');
        foreach ($args['otpInfo'] as $name => $value) {
            $metaDataElement = $otpInfoXML->addChild('meta-data');
            $metaDataElement->addAttribute('name', $name);
            $metaDataElement->addAttribute('value', $value);
        }

        $xmlData = $otpInfoXML->asXML();
        if (!empty($args['prettyPrint'])) {
            $dom = dom_import_simplexml($otpInfoXML)->ownerDocument;
            $dom->formatOutput = true;
            $xmlData = $dom->saveXML();
        }

        return ['filename' => $args['filename'], 'content' => $xmlData];
    }

    public static function download(array $args)
    {
        ValidatorModel::notEmpty($args, ['connection_data', 'documentId']);
        ValidatorModel::arrayType($args, ['connection_data']);
        ValidatorModel::intVal($args, ['documentId']);

        $curlReturn = CurlModel::exec([
            'method'        => 'GET',
            'url'           => $args['connection_data']['apiUri'] . '/documents/v2/' . $args['documentId'] . '/download',
            'options'       => [
                CURLOPT_SSLCERT       => $args['connection_data']['certPath'],
                CURLOPT_SSLCERTPASSWD => AuthenticationModel::decrypt(['cryptedPassword' => $args['connection_data']['certPass']]),
                CURLOPT_SSLCERTTYPE   => $args['connection_data']['certType'],
            ],
            'fileResponse'  => true
        ]);

        if ($curlReturn['code'] != 200) {
            return ['errors' => "Raw Error : {$curlReturn['raw']}"];
        } elseif (!empty($curlReturn['response']['userFriendlyMessage']) && $curlReturn['response']['userFriendlyMessage'] != 'Erreur') {
            return ['errors' => $curlReturn['response']['userFriendlyMessage']];
        } elseif (!empty($curlReturn['response']['developerMessage'])) {
            return ['errors' => $curlReturn['response']['developerMessage']];
        }

        return ['b64FileContent' => base64_encode($curlReturn['response'])];
    }

    public static function getRefusalMessage(array $args)
    {
        ValidatorModel::notEmpty($args, ['connection_data', 'documentId']);
        ValidatorModel::arrayType($args, ['connection_data']);
        ValidatorModel::intVal($args, ['documentId']);

        $curlReturn = CurlModel::exec([
            'url'           => $args['connection_data']['apiUri'] . '/documents/v2/' . $args['documentId'] . '/comments/refusal',
            'method'        => 'GET',
            'options'       => [
                CURLOPT_SSLCERT       => $args['connection_data']['certPath'],
                CURLOPT_SSLCERTPASSWD => AuthenticationModel::decrypt(['cryptedPassword' => $args['connection_data']['certPass']]),
                CURLOPT_SSLCERTTYPE   => $args['connection_data']['certType']
            ]
        ]);

        if ($curlReturn['code'] != 200) {
            return ['errors' => $curlReturn['raw']];
        } elseif (!empty($curlReturn['response']['errorCode'])) {
            if ($curlReturn['response']['errorCode'] == 102 && $curlReturn['response']['developerMessage'] == "Internal error: Le document n'a pas de motif de refus ou rejet") {
                return ['response' => ''];
            } elseif (!empty($curlReturn['response']['userFriendlyMessage']) && $curlReturn['response']['userFriendlyMessage'] != 'Erreur') {
                return ['errors' => $curlReturn['response']['userFriendlyMessage']];
            } else {
                return ['errors' => $curlReturn['response']['developerMessage']];
            }
        }

        return ['response' => $curlReturn['response']['comment']];
    }

    /**
     * The “delete” method allows you to delete a document.
     * The document must be in the correct state to be deleted.
     * Please note that this deletion is permanent.
     * The list of status allowing a document to be deleted are :
     *  - Préparé
     *  - Refusé
     *  - Rejeté
     */
    public static function delete(array $args)
    {
        ValidatorModel::notEmpty($args, ['connection_data', 'documentId']);
        ValidatorModel::arrayType($args, ['connection_data']);
        ValidatorModel::intVal($args, ['documentId']);

        $curlReturn = CurlModel::exec([
            'method'        => 'DELETE',
            'url'           => $args['connection_data']['apiUri'] . '/documents/v2/' . $args['documentId'],
            'options'       => [
                CURLOPT_SSLCERT       => $args['connection_data']['certPath'],
                CURLOPT_SSLCERTPASSWD => AuthenticationModel::decrypt(['cryptedPassword' => $args['connection_data']['certPass']]),
                CURLOPT_SSLCERTTYPE   => $args['connection_data']['certType'],
            ]
        ]);

        if (!in_array($curlReturn['code'], [200, 204])) {
            return ['errors' => "Raw Error : {$curlReturn['raw']}"];
        } elseif (!empty($curlReturn['response']['userFriendlyMessage']) && $curlReturn['response']['userFriendlyMessage'] != 'Erreur') {
            return ['errors' => $curlReturn['response']['userFriendlyMessage']];
        } elseif (!empty($curlReturn['response']['developerMessage'])) {
            return ['errors' => $curlReturn['response']['developerMessage']];
        }

        return true;
    }

    private static function generateToken(array $args)
    {
        ValidatorModel::notEmpty($args, ['payload']);
        ValidatorModel::arrayType($args, ['payload']);

        return AuthenticationModel::generateToken($args['payload'], CoreConfigModel::getEncryptKey());
    }

    private static function checkToken(array $args)
    {
        if (!Validator::stringType()->notEmpty()->validate($args['token'] ?? null)) {
            return ['errors' => 'token is empty or not a string'];
        }

        try {
            $jwt = AuthenticationModel::decodeToken($args['token'], CoreConfigModel::getEncryptKey());
        } catch (\Exception $e) {
            return ['errors' => 'FastParapheur access token is invalid'];
        }

        if (empty($jwt['documentId']) || empty($jwt['workflowId'])) {
            return ['errors' => 'FastParapheur access token is invalid'];
        }

        $workflowExternalInformations = WorkflowExternalInformationModel::getByWorkflowId(['select' => ['informations'], 'workflowId' => $jwt['workflowId']]);
        if (empty($workflowExternalInformations)) {
            return ['errors' => 'Workflow external informations does not exist'];
        }
        $informations = json_decode($workflowExternalInformations['informations'], true);

        if (!empty($informations['token']) && $args['token'] != $informations['token']) {
            return ['errors' => 'FastParapheur access token is invalid'];
        }

        return ['documentId' => $jwt['documentId'], 'workflowId' => $jwt['workflowId']];
    }

    public static function getProof(array $args)
    {
        ValidatorModel::notEmpty($args, ['connection_data', 'documentId']);
        ValidatorModel::arrayType($args, ['connection_data']);
        ValidatorModel::intVal($args, ['documentId']);

        $curlReturn = CurlModel::exec([
            'url'           => $args['connection_data']['apiUri'] . '/documents/v2/' . $args['documentId'] . '/getFdc',
            'method'        => 'GET',
            'options'       => [
                CURLOPT_SSLCERT       => $args['connection_data']['certPath'],
                CURLOPT_SSLCERTPASSWD => AuthenticationModel::decrypt(['cryptedPassword' => $args['connection_data']['certPass']]),
                CURLOPT_SSLCERTTYPE   => $args['connection_data']['certType'],
            ],
            'fileResponse'  => true
        ]);

        if ($curlReturn['code'] != 200) {
            return ['errors' => "Raw Error : {$curlReturn['raw']}"];
        } elseif (!empty($curlReturn['response']['userFriendlyMessage']) && $curlReturn['response']['userFriendlyMessage'] != 'Erreur') {
            return ['errors' => $curlReturn['response']['userFriendlyMessage']];
        } elseif (!empty($curlReturn['response']['developerMessage'])) {
            return ['errors' => $curlReturn['response']['developerMessage']];
        }
        return ['response' => $curlReturn['response']];
    }

    public static function getProofByWorkflowId(array $args)
    {
        ValidatorModel::notEmpty($args, ['workflowId']);
        ValidatorModel::intVal($args, ['workflowId']);

        $workflowExternalInformations = WorkflowExternalInformationModel::getByWorkflowId(['select' => ['informations', 'external_signatory_book_id'], 'workflowId' => $args['workflowId']]);
        if (empty($workflowExternalInformations)) {
            return ['errors' => 'Workflow does not exist'];
        }

        $informations = json_decode($workflowExternalInformations['informations'], true);

        $externalSB = ExternalSignatoryBookModel::getById(['select' => ['*'], 'id' => $workflowExternalInformations['external_signatory_book_id']]);
        if (empty($externalSB)) {
            return ['errors' => 'No external signatory book configuration found'];
        } elseif ($externalSB['type'] != 'fast') {
            return ['errors' => 'External signatory book configuration is invalid'];
        }
        $externalSB['connection_data'] = json_decode($externalSB['connection_data'], true);

        $proofResponse = FastOTPController::getProof([
            'connection_data'   => $externalSB['connection_data'],
            'documentId'        => $informations['fastDocId'],
        ]);
        if (!empty($proofResponse['errors'])) {
            return ['errors' => $proofResponse['errors']];
        }

        return ['data' => $proofResponse['response'], 'filename' => 'ficheDeCirculation.pdf'];
    }
}
