<?php

/**
* Copyright Maarch since 2008 under license.
* See LICENSE.txt file at the root folder for more details.
* This file is part of Maarch software.
*
*/

/**
* @brief Workflow Model
* @author dev@maarch.org
*/

namespace Workflow\models;

use SrcCore\models\ValidatorModel;
use SrcCore\models\DatabaseModel;

class WorkflowModel
{
    public static function get(array $args)
    {
        ValidatorModel::notEmpty($args, ['select']);
        ValidatorModel::arrayType($args, ['select', 'where', 'data', 'orderBy', 'groupBy']);
        ValidatorModel::intType($args, ['limit', 'offset']);

        $workflows = DatabaseModel::select([
            'select'    => $args['select'],
            'table'     => ['workflows'],
            'where'     => empty($args['where']) ? [] : $args['where'],
            'data'      => empty($args['data']) ? [] : $args['data'],
            'orderBy'   => empty($args['orderBy']) ? [] : $args['orderBy'],
            'groupBy'   => empty($args['groupBy']) ? [] : $args['groupBy'],
            'offset'    => empty($args['offset']) ? 0 : $args['offset'],
            'limit'     => empty($args['limit']) ? 0 : $args['limit'],
        ]);

        return $workflows;
    }

    public static function getById(array $args)
    {
        ValidatorModel::notEmpty($args, ['select', 'id']);
        ValidatorModel::intVal($args, ['id']);
        ValidatorModel::arrayType($args, ['select']);

        $workflow = DatabaseModel::select([
            'select'    => $args['select'],
            'table'     => ['workflows'],
            'where'     => ['id = ?'],
            'data'      => [$args['id']]
        ]);

        if (empty($workflow[0])) {
            return [];
        }

        return $workflow[0];
    }

    public static function getByDocumentId(array $args)
    {
        ValidatorModel::notEmpty($args, ['select', 'documentId']);
        ValidatorModel::intVal($args, ['documentId']);
        ValidatorModel::arrayType($args, ['select', 'orderBy']);

        $workflow = DatabaseModel::select([
            'select'    => $args['select'],
            'table'     => ['workflows'],
            'where'     => ['main_document_id = ?'],
            'data'      => [$args['documentId']],
            'orderBy'   => empty($args['orderBy']) ? [] : $args['orderBy'],
        ]);

        if (empty($workflow)) {
            return [];
        }

        return $workflow;
    }

    /**
     * returns the current step of a given document’s workflow
     *
     * a step is considered to be the current one if there are no processed steps after it.
     */
    public static function getCurrentStep(array $args)
    {
        ValidatorModel::notEmpty($args, ['select', 'documentId']);
        ValidatorModel::intVal($args, ['documentId']);
        ValidatorModel::arrayType($args, ['select']);

        $currentStep = DatabaseModel::select([
            'select'    => $args['select'],
            'table'     => ['workflows'],
            'where'     => ['main_document_id = ?', 'process_date is null', 'status is null'],
            'data'      => [$args['documentId']],
            'orderBy'   => ['"order"'],
            'limit'     => 1,
        ]);

        if (empty($currentStep[0])) {
            return [];
        }

        return $currentStep[0];
    }

    /**
     * returns whether a workflow step identified by its id is the last in its own workflow
     *
     * does NOT depend on the workflow state: it may be processed or not, it only has to be the last one scheduled.
     */
    public static function isLastStep(array $args)
    {
        ValidatorModel::notEmpty($args, ['id']);
        ValidatorModel::intVal($args, ['id']);

        $return = DatabaseModel::select([
            'select'    => [1],
            'table'     => ['workflows ws1', 'workflows ws2'],
            'left_join' => ['ws1.main_document_id = ws2.main_document_id'],
            'where'     => ['ws1.id = ?', 'ws2.order > ws1.order'],
            'data'      => [$args['id']],
            'limit'     => 1
        ]);

        return empty($return);
    }

    public static function getWorkflowWithExternalInfo(array $aArgs)
    {
        ValidatorModel::arrayType($aArgs, ['select', 'orderBy', 'where', 'data']);

        $workflows = DatabaseModel::select([
            'select'    => empty($aArgs['select']) ? ['*'] : $aArgs['select'],
            'table'     => ['workflows w', 'workflows_external_informations wei'],
            'left_join' => ['wei.workflow_id = w.id'],
            'where'     => empty($aArgs['where']) ? [] : $aArgs['where'],
            'data'      => empty($aArgs['data']) ? [] : $aArgs['data'],
            'orderBy'   => empty($aArgs['orderBy']) ? [] : $aArgs['orderBy'],
            'limit'     => empty($aArgs['limit']) ? 0 : $aArgs['limit']
        ]);

        return $workflows;
    }

    public static function create(array $args)
    {
        ValidatorModel::notEmpty($args, ['mainDocumentId', 'mode', 'order', 'signatureMode']);
        ValidatorModel::intVal($args, ['userId', 'mainDocumentId', 'order']);
        ValidatorModel::stringType($args, ['mode', 'signatureMode', 'signaturePositions', 'datePositions']);

        $nextSequenceId = DatabaseModel::getNextSequenceValue(['sequenceId' => 'workflows_id_seq']);

        DatabaseModel::insert([
            'table'         => 'workflows',
            'columnsValues' => [
                'id'                    => $nextSequenceId,
                'user_id'               => $args['userId'],
                'main_document_id'      => $args['mainDocumentId'],
                'mode'                  => $args['mode'],
                'signature_mode'        => $args['signatureMode'],
                'signature_positions'   => $args['signaturePositions'],
                'date_positions'        => $args['datePositions'],
                '"order"'               => $args['order']
            ]
        ]);

        return $nextSequenceId;
    }

    public static function update(array $args)
    {
        ValidatorModel::notEmpty($args, ['set', 'where', 'data']);
        ValidatorModel::arrayType($args, ['set', 'where', 'data']);

        DatabaseModel::update([
            'table' => 'workflows',
            'set'   => $args['set'],
            'where' => $args['where'],
            'data'  => $args['data']
        ]);

        return true;
    }

    public static function delete(array $args)
    {
        ValidatorModel::notEmpty($args, ['where', 'data']);
        ValidatorModel::arrayType($args, ['where', 'data']);

        DatabaseModel::delete([
            'table' => 'workflows',
            'where' => $args['where'],
            'data'  => $args['data']
        ]);

        return true;
    }
}
