import { Component, OnInit, TemplateRef, ViewChild } from '@angular/core';
import { SignaturesContentService } from '../../service/signatures.service';
import { NotificationService } from '../../service/notification.service';
import { HttpClient } from '@angular/common/http';
import { MatDialog } from '@angular/material/dialog';
import { MatSort, Sort } from '@angular/material/sort';
import { TranslateService } from '@ngx-translate/core';
import { map, finalize, tap, catchError } from 'rxjs/operators';
import { LatinisePipe } from 'ngx-pipes';
import { AuthService } from '../../service/auth.service';
import { AlertController } from '@ionic/angular';
import { FunctionsService } from '../../service/functions.service';
import { TitleCasePipe } from '@angular/common';
import { of } from 'rxjs';


export interface User {
    id: string;
    firstname: string;
    lastname: string;
    login: string;
    email: string;
    subtitute: boolean;
}

@Component({
    templateUrl: 'notifications-list.component.html',
    styleUrls: ['../administration.scss', 'notifications-list.component.scss'],
    providers: [TitleCasePipe]
})

export class NotificationsListComponent {

    scheduleList: User[] = [];
    sortedData: any[];
    displayedColumns: string[];
    loading: boolean = true;
    batchMasterInitilized: boolean = true;
    applicationUrl: string = '';

    constructor(
        public http: HttpClient,
        private translate: TranslateService,
        private latinisePipe: LatinisePipe,
        public dialog: MatDialog,
        public signaturesService: SignaturesContentService,
        public functionsService: FunctionsService,
        public notificationService: NotificationService,
        public authService: AuthService,
        public alertController: AlertController,
        private titleCasePipe: TitleCasePipe
    ) {

        if (this.signaturesService.mobileMode) {
            this.displayedColumns = ['label', 'type', 'frequency', 'actions'];
        } else {
            this.displayedColumns = ['label', 'type', 'frequency', 'actions'];
        }
    }

    applyFilter(filterValue: string) {
        filterValue = this.latinisePipe.transform(filterValue.toLowerCase());

        this.sortedData = this.scheduleList.filter(
            (option: any) => {
                let state = false;
                this.displayedColumns.forEach(element => {
                    if (option[element] && this.latinisePipe.transform(option[element].toLowerCase()).includes(filterValue)) {
                        state = true;
                    }
                });
                return state;
            }
        );
    }

    ionViewWillEnter() {
        this.checkInitializedBatchMaster();
        this.getApplicationUrl();
        this.http.get('../rest/schedule')
            .pipe(
                map((data: any) => data.notificationsSchedule.map((item: any) => ({
                    ...item,
                    type: this.translate.instant('lang.notification_' + item.type),
                }))),
                finalize(() => this.loading = false)
            )
            .subscribe({
                next: data => {
                    this.scheduleList = data;
                    this.formatFrequency();

                    this.sortedData = this.scheduleList.slice();
                },
            });
    }

    checkInitializedBatchMaster() {
        this.http.get('../rest/schedule/initialization')
            .pipe(
                tap((data: any) => {
                    this.batchMasterInitilized = data.schedulerInitialized;
                }),
                catchError((err: any) => {
                    this.notificationService.handleErrors(err);
                    return of(false);
                })
            ).subscribe();
    }

    initalizeBatchMaster() {
        this.http.post('../rest/schedule/initialization', {})
            .pipe(
                tap((data: any) => {
                    this.batchMasterInitilized = true;
                    this.notificationService.success('lang.schedulerInitialized');
                }),
                catchError((err: any) => {
                    this.notificationService.handleErrors(err);
                    return of(false);
                })
            ).subscribe();
    }

    formatFrequency() {
        this.scheduleList.forEach((notif: any) => {
            let frequencyLabel = '';
            if (notif.months[0] === '*' && notif.daysOfMonth[0] === '*') {
                frequencyLabel = this.translate.instant('lang.frequencyDaily', { 0: notif.daysOfWeek.map((item: any) => this.functionsService.getDayOfWeekString(item)).join(', ') });
            } else if (notif.daysOfWeek[0] === '*') {
                frequencyLabel = this.translate.instant('lang.frquencyMonthly', { months: notif.months.map((item: any) => this.functionsService.getMonthString(item)).join(', '), daysOfMonth: notif.daysOfMonth.join(', ') });
            }

            if (!this.functionsService.empty(notif.frequency)) {
                frequencyLabel += '<br>' + this.translate.instant('lang.frequencyHourly', { frequency: notif.frequency + ' ' + this.translate.instant('lang.' + notif.frequencyMode), startTime: notif.startTime, endTime: notif.endTime });
            } else {
                frequencyLabel += '<br>' + this.titleCasePipe.transform(this.translate.instant('lang.atRange')) + ' ' + notif.startTime;
            }
            notif.frequency = frequencyLabel;
        });
    }


    async delete(scheduleToDelete: User) {
        const alert = await this.alertController.create({
            // cssClass: 'custom-alert-danger',
            header: this.translate.instant('lang.confirmMsg'),
            buttons: [
                {
                    text: this.translate.instant('lang.no'),
                    role: 'cancel',
                    cssClass: 'secondary',
                    handler: () => { }
                },
                {
                    text: this.translate.instant('lang.yes'),
                    handler: () => {
                        this.http.delete('../rest/schedule/' + scheduleToDelete.id)
                            .pipe(
                                finalize(() => this.loading = false)
                            )
                            .subscribe({
                                next: data => {
                                    const indexToDelete = this.scheduleList.findIndex(user => user.id === scheduleToDelete.id);

                                    this.scheduleList.splice(indexToDelete, 1);
                                    this.sortedData = this.scheduleList.slice();
                                    this.notificationService.success('lang.scheduleDeleted');
                                },
                                error: err => {
                                    this.notificationService.handleErrors(err);
                                }
                            });
                    }
                }
            ]
        });

        await alert.present();
    }

    sortData(sort: Sort) {
        const data = this.scheduleList.slice();
        if (!sort.active || sort.direction === '') {
            this.sortedData = data;
            return;
        }

        this.sortedData = data.sort((a, b) => {
            const isAsc = sort.direction === 'asc';
            return compare(a[sort.active], b[sort.active], isAsc);
        });
    }


    getApplicationUrl() {
        this.http.get('../rest/authenticationInformations').pipe(
            tap((data: any) => {
                this.applicationUrl = data.applicationUrl;
                this.loading = false;
            }),
            catchError((err: any) => {
                this.notificationService.error(err);
                return of(false);
            })
        ).subscribe();
    }
}

function compare(a: number | string, b: number | string, isAsc: boolean) {
    return (a < b ? -1 : 1) * (isAsc ? 1 : -1);
}
