import { Component, OnInit, ViewChild } from '@angular/core';
import { SignaturesContentService } from '../../service/signatures.service';
import { NotificationService } from '../../service/notification.service';
import { HttpClient, HttpHeaders } from '@angular/common/http';
import { MatDialog } from '@angular/material/dialog';
import {map, finalize, tap, catchError} from 'rxjs/operators';
import { ActivatedRoute, Router } from '@angular/router';
import { ConfirmComponent } from '../../plugins/confirm.component';
import { TranslateService } from '@ngx-translate/core';
import { AuthService } from '../../service/auth.service';
import { of } from 'rxjs';
import { AlertController, ModalController } from '@ionic/angular';
import { GroupModalComponent } from '../group/modal/group-modal.component';


export interface User {
    id: string;
    firstname: string;
    lastname: string;
    login: string;
    email: string;
    phone: string;
    picture: string;
    isRest: boolean;
    signatureModes: string[];
    canSendActivationNotification: boolean;
}

@Component({
    selector: 'app-administration-user',
    templateUrl: 'user.component.html',
    styleUrls: ['../administration.scss', 'user.component.scss'],
})

export class UserComponent implements OnInit {

    creationMode: boolean = true;
    loading: boolean = true;
    user: User = {
        id: '',
        firstname: '',
        lastname: '',
        login: '',
        email: '',
        phone: '',
        picture: '',
        isRest: false,
        signatureModes: ['stamp'],
        canSendActivationNotification: false
    };
    userClone: User;
    title: string = '';
    hideCurrentPassword: Boolean = true;
    hideNewPassword: Boolean = true;
    hideNewPasswordConfirm: Boolean = true;
    currentTool = 'info';

    // HANDLE PASSWORD
    passwordRules: any = {
        minLength: { enabled: false, value: 0 },
        complexityUpper: { enabled: false, value: 0 },
        complexityNumber: { enabled: false, value: 0 },
        complexitySpecial: { enabled: false, value: 0 },
        renewal: { enabled: false, value: 0 },
        historyLastUse: { enabled: false, value: 0 },
    };
    passwordRest: any = {
        newPassword: '',
        passwordConfirmation: ''
    };

    ruleText = '';
    otherRuleText = '';

    showPassword = false;
    handlePassword: any = {
        error: false,
        errorMsg: ''
    };

    hasGroup: boolean = false;
    userGroupsClone: any;
    manageableGroups: any;

    constructor(
        public http: HttpClient,
        private translate: TranslateService,
        private route: ActivatedRoute,
        private router: Router,
        public signaturesService: SignaturesContentService,
        public notificationService: NotificationService,
        public dialog: MatDialog,
        public authService: AuthService,
        public alertController: AlertController,
        public modalController: ModalController,
    ) {
    }

    async ngOnInit(): Promise<void> {
        this.route.params.subscribe(async (params: any) => {
            if (params['id'] === undefined) {
                this.creationMode = true;
                this.title = this.translate.instant('lang.userCreation');
                this.user = {
                    id: '',
                    firstname: '',
                    lastname: '',
                    login: '',
                    email: '',
                    phone: '',
                    picture: '',
                    signatureModes: ['stamp', 'note'],
                    isRest: false,
                    canSendActivationNotification: false
                };
                this.loading = false;
                await this.isCreationMode();
            } else {
                this.isUpdateMode(params['id']);
                this.loading = false;
            }
        });
    }

    async isCreationMode() {
        this.creationMode = true;
        this.title = this.translate.instant('lang.userCreation');
        this.user = {
            id: '',
            firstname: '',
            lastname: '',
            login: '',
            email: '',
            phone: '',
            picture: '',
            signatureModes: ['stamp'],
            isRest: false,
            canSendActivationNotification: false
        };
        await this.getManageableGroups();
        this.loading = false;
    }

    isUpdateMode(id: any) {
        this.creationMode = false;
        this.http.get(`../rest/users/${id}`)
            .pipe(
                map((data: any) => data.user)
            )
            .subscribe({
                next: async data => {
                    this.user = data;
                    this.hasGroup = data.groups.length > 0;
                    this.userGroupsClone = JSON.parse(JSON.stringify(data.groups));
                    await this.getManageableGroups();
                    this.userClone = JSON.parse(JSON.stringify(this.user));
                    this.title = this.user.firstname + ' ' + this.user.lastname;
                    if (this.user.isRest) {
                        this.getPassRules({ detail: {
                            checked: true
                        } });
                    }
                },
            });
    }

    canValidate() {
        if (this.user.isRest && this.passwordRest.newPassword !== '' && (this.handlePassword.error || this.passwordRest.passwordConfirmation !== this.passwordRest.newPassword)) {
            return false;
        } else if (JSON.stringify(this.user) === JSON.stringify(this.userClone) && this.passwordRest.newPassword === '') {
            return false;
        } else {
            return true;
        }
    }

    onSubmit() {
        if (this.creationMode) {
            this.createUser();
        } else {
            this.modifyUser();
        }
    }

    modifyUser() {
        this.loading = true;
        const objToSend = {...this.user, groups: this.user['groups'].map((group: any) => group.id)};
        this.http.put('../rest/users/' + this.user.id, objToSend)
            .pipe(
                finalize(() => this.loading = false),
                tap(() => {
                    if (this.authService.user.id === this.user.id) {
                        this.authService.updateUserInfoWithTokenRefresh();
                    }
                    if (this.passwordRest.newPassword !== '') {
                        this.updateRestUser();
                    }
                    this.router.navigate(['/administration/users/' + this.user.id]);
                    this.notificationService.success('lang.userUpdated');
                }),
                catchError((err: any) => {
                    this.notificationService.handleErrors(err);
                    return of(false);
                })
            )
            .subscribe();
    }

    updateRestUser() {
        const headers = new HttpHeaders({
            'Authorization': 'Bearer ' + this.authService.getToken()
        });
        this.http.put('../rest/users/' + this.user.id + '/password', this.passwordRest, { headers: headers })
            .subscribe(() => {
                this.passwordRest.newPassword = '';
                this.passwordRest.passwordConfirmation = '';
            }, (err) => {
                this.notificationService.handleErrors(err);
            });
    }

    createUser() {
        this.loading = true;
        this.http.post('../rest/users', this.user)
            .pipe(
                finalize(() => this.loading = false),
                tap((data: any) => {
                    if (this.user.isRest) {
                        this.user.id = data.id;
                        this.updateRestUser();
                    }
                    this.router.navigate([`/administration/users/${data.id}`]);
                    this.notificationService.success('lang.userAdded');
                }),
                catchError((err: any) => {
                    this.notificationService.handleErrors(err);
                    return of(false);
                })
            )
            .subscribe();
    }

    delete() {
        const dialogRef = this.dialog.open(ConfirmComponent, { autoFocus: false, data: { mode: '', title: 'lang.confirmMsg', msg: '' } });

        dialogRef.afterClosed().subscribe(result => {
            if (result === 'yes') {
                this.loading = true;
                this.http.delete('../rest/users/' + this.user.id)
                    .pipe(
                        finalize(() => this.loading = false)
                    )
                    .subscribe({
                        next: () => {
                            this.router.navigate(['/administration/users']);
                            this.notificationService.success('lang.userDeleted');
                        },
                    });
            }
        });
    }

    cancel() {
        this.router.navigate(['/administration/users']);
    }

    getPassRules(ev: any) {
        if (ev.detail.checked) {
            this.handlePassword.error = false;
            this.handlePassword.errorMsg = '';

            this.http.get('../rest/passwordRules')
                .subscribe((data: any) => {
                    const ruleTextArr: String[] = [];
                    const otherRuleTextArr: String[] = [];

                    data.rules.forEach((rule: any) => {
                        if (rule.label === 'minLength') {
                            this.passwordRules.minLength.enabled = rule.enabled;
                            this.passwordRules.minLength.value = rule.value;
                            if (rule.enabled) {
                                this.translate.get('lang.minLengthChar', { charLength: rule.value }).subscribe((res: string) => {
                                    ruleTextArr.push(res);
                                });
                            }

                        } else if (rule.label === 'complexityUpper') {
                            this.passwordRules.complexityUpper.enabled = rule.enabled;
                            this.passwordRules.complexityUpper.value = rule.value;
                            if (rule.enabled) {
                                ruleTextArr.push('lang.upperRequired');
                            }

                        } else if (rule.label === 'complexityNumber') {
                            this.passwordRules.complexityNumber.enabled = rule.enabled;
                            this.passwordRules.complexityNumber.value = rule.value;
                            if (rule.enabled) {
                                ruleTextArr.push('lang.numberRequired');
                            }

                        } else if (rule.label === 'complexitySpecial') {
                            this.passwordRules.complexitySpecial.enabled = rule.enabled;
                            this.passwordRules.complexitySpecial.value = rule.value;
                            if (rule.enabled) {
                                ruleTextArr.push('lang.specialCharRequired');
                            }
                        } else if (rule.label === 'renewal') {
                            this.passwordRules.renewal.enabled = rule.enabled;
                            this.passwordRules.renewal.value = rule.value;
                            if (rule.enabled) {
                                this.translate.get('lang.renewalInfo', { time: rule.value }).subscribe((res: string) => {
                                    otherRuleTextArr.push(res);
                                });
                            }
                        } else if (rule.label === 'historyLastUse') {
                            this.passwordRules.historyLastUse.enabled = rule.enabled;
                            this.passwordRules.historyLastUse.value = rule.value;
                            if (rule.enabled) {
                                this.translate.get('lang.historyUseInfo', { countPwd: rule.value }).subscribe((res: string) => {
                                    otherRuleTextArr.push(res);
                                });
                            }
                        }

                    });
                    this.ruleText = ruleTextArr.join(', ');
                    this.otherRuleText = otherRuleTextArr.join('<br/>');
                }, (err) => {
                    this.notificationService.handleErrors(err);
                });
        }
    }

    checkPasswordValidity(password: string) {
        this.handlePassword.error = true;

        if (!password.match(/[A-Z]/g) && this.passwordRules.complexityUpper.enabled) {
            this.handlePassword.errorMsg = 'lang.upperRequired';
        } else if (!password.match(/[0-9]/g) && this.passwordRules.complexityNumber.enabled) {
            this.handlePassword.errorMsg = 'lang.numberRequired';
        } else if (!password.match(/[^A-Za-z0-9]/g) && this.passwordRules.complexitySpecial.enabled) {
            this.handlePassword.errorMsg = 'lang.specialCharRequired';
        } else if (password.length < this.passwordRules.minLength.value && this.passwordRules.minLength.enabled) {
            this.translate.get('lang.minLengthChar', { charLength: this.passwordRules.minLength.value }).subscribe((res: string) => {
                this.handlePassword.errorMsg = res;
            });
        } else {
            this.handlePassword.error = false;
            this.handlePassword.errorMsg = '';
        }
    }

    toggleSignMode(signMode: any, state: boolean) {
        if (state) {
            if (this.user.signatureModes.indexOf(signMode.id) === -1) {
                this.user.signatureModes.push(signMode.id);
            }
        } else {
            this.user.signatureModes = this.user.signatureModes.filter((item: any) => item !== signMode.id);
        }
    }

    initTab(val: any) {
        this.currentTool = val;
    }

    async resendActivationNotification() {
        const alert = await this.alertController.create({
            header: this.translate.instant('lang.confirmMsg'),
            buttons: [
                {
                    text: this.translate.instant('lang.no'),
                    role: 'cancel',
                    cssClass: 'secondary',
                    handler: () => { }
                },
                {
                    text: this.translate.instant('lang.yes'),
                    handler: () => {
                        this.http.put('../rest/users/' + this.user.id + '/accountActivationNotification', {})
                            .pipe(
                                tap(() => {
                                    this.notificationService.success(this.translate.instant('lang.activationNotificationSend'));
                                }),
                                catchError((err: any) => {
                                    this.notificationService.handleErrors(err);
                                    return of(false);
                                })
                            )
                            .subscribe();
                    }
                }
            ]
        });

        await alert.present();
    }

    getManageableGroups() {
        return new Promise((resolve) => {
            this.http.get('../rest/manageableGroups').pipe(
                tap((data: any) => {
                    this.manageableGroups = JSON.parse(JSON.stringify(data.groups));
                    if (this.hasGroup) {
                        const groupIds: number[] = this.userGroupsClone.map((group: any) => group.id);
                        this.userGroupsClone = JSON.parse(JSON.stringify(data.groups.filter((group: any) => groupIds.indexOf(group.id) === -1)));
                    } else  {
                        this.userGroupsClone = JSON.parse(JSON.stringify(data.groups));
                    }
                    resolve(true);
                }),
                finalize(() => this.loading = false),
                catchError((err: any) => {
                    this.notificationService.handleErrors(err);
                    return of(false);
                })
            ).subscribe();
        });
    }

    isManageableGroup(groupId: any) {
        return this.manageableGroups.find((userGroup: any) => userGroup.id === groupId) !== undefined;
    }

    async unlinkGroup(group: any) {
        const cannotManage: boolean = this.user['groups'].length === 1 && this.user.id === this.authService.user.id && this.authService.user.administrativePrivileges.find((privilege: any) => privilege.id === 'manage_users') !== undefined;
        const alert = await this.alertController.create({
            header: this.translate.instant('lang.confirmMsg'),
            message: cannotManage ? this.translate.instant('lang.groupWarnMsg') : '',
            buttons: [
                {
                    text: this.translate.instant('lang.no'),
                    role: 'cancel',
                    cssClass: 'secondary',
                    handler: () => { }
                },
                {
                    text: this.translate.instant('lang.yes'),
                    handler: () => {
                        this.http.delete(`../rest/groups/${group.id}/users/${this.user.id}`, {})
                            .pipe(
                                tap(() => {
                                    this.isUpdateMode(this.user.id);
                                    this.userGroupsClone.push(group);
                                }),
                                finalize(() => this.loading = false)
                            )
                            .subscribe({
                                next: data => {
                                    const indexToDelete = this.user['groups'].findIndex((item: any) => item.id === group.id);
                                    this.user['groups'].splice(indexToDelete, 1);
                                    this.notificationService.success('lang.groupDeleted');
                                    this.authService.updateUserInfoWithTokenRefresh();
                                    this.http.get(`../rest/users/${this.user.id}`).pipe(
                                        tap(() => {}),
                                        catchError((err: any) => {
                                            if (err.error.errors === 'User out of perimeter') {
                                                this.router.navigate(['/administration/users']);
                                            }
                                            return of(false);
                                        })
                                    ).subscribe();
                                },
                                error: err => {
                                    this.notificationService.handleErrors(err);
                                }
                            });
                    }
                }
            ]
        });
        await alert.present();
    }

    async openGroupsList() {
        const modal = await this.modalController.create({
            component: GroupModalComponent,
            componentProps: {
                groups: this.userGroupsClone,
                editUser: true
            }
        });
        await modal.present();
        const { data } = await modal.onWillDismiss();
        if (data !== undefined) {
            this.user['groups'] = this.user['groups'].concat(data.filter((group: any) => group.checked).map((item: any) => ({
                id: item.id,
                label: item.label
            })));
            const groupIds: number[] = this.user['groups'].map((group: any) => group.id);
            this.userGroupsClone = this.userGroupsClone.filter((group: any) => groupIds.indexOf(group.id) === -1);
            this.modifyUser();
        }
    }
}
