import { Component, AfterViewInit, OnInit } from '@angular/core';
import { MatDialog } from '@angular/material/dialog';
import { HttpClient } from '@angular/common/http';
import { ActivatedRoute, Router } from '@angular/router';
import { SignaturesContentService } from '../service/signatures.service';
import { NotificationService } from '../service/notification.service';
import { environment } from '../../core/environments/environment';
import { Validators, UntypedFormGroup, UntypedFormBuilder } from '@angular/forms';
import { AuthService } from '../service/auth.service';
import { tap, catchError } from 'rxjs/operators';
import { of } from 'rxjs';
import { LoadingController, MenuController } from '@ionic/angular';
import { TranslateService } from '@ngx-translate/core';
import { FunctionsService } from '../service/functions.service';
import { OtpService } from '../service/otp-connectors/otp-generator.service';

@Component({
    templateUrl: 'login.component.html',
    styleUrls: ['login.component.scss'],
})
export class LoginComponent implements OnInit, AfterViewInit {

    loginForm: UntypedFormGroup;

    loading: boolean = false;
    showForm: boolean = false;
    environment: any;
    commitHash: any = null;
    hidePassword: boolean = true;
    applicationVersion: string = '';


    constructor(
        private http: HttpClient,
        private router: Router,
        public authService: AuthService,
        private signaturesService: SignaturesContentService,
        private notificationService: NotificationService,
        public dialog: MatDialog,
        private formBuilder: UntypedFormBuilder,
        public loadingController: LoadingController,
        public functions: FunctionsService,
        private translate: TranslateService,
        private menu: MenuController,
        private _activatedRoute: ActivatedRoute,
        private otpService: OtpService
    ) {
        this._activatedRoute.url.subscribe((param: any[]) => {
            const urlParams: any = {
                token: null,
                status: null
            };
            if (!this.functions.empty(param)) {
                param.forEach((item: any, index: number) => {
                    if (item.path === 'token') {
                        urlParams['token'] = param[index + 1].path;
                    }
                });
            }

            this._activatedRoute.queryParams.subscribe(queryParam => {
                if (!this.functions.empty(queryParam)) {
                    urlParams['status'] = queryParam.return;
                }
            });

            if (!this.functions.empty(urlParams['status']) && !this.functions.empty(urlParams['token'])) {
                this.otpService.getDocument(urlParams);
            }
        });
    }

    async ngOnInit() {
        const appVersion: string = environment.VERSION.split('.')[0] + '.' + environment.VERSION.split('.')[1];
        if (!this.functions.empty(environment.VERSION.split(' ')[1])) {
            const mode: string = (environment.VERSION.split(' ')[1]).concat(' ', environment.VERSION.split(' ')[2]);
            this.applicationVersion = appVersion.concat(' ', mode);
        } else {
            this.applicationVersion = appVersion;
        }

        this.loginForm = this.formBuilder.group({
            login: [null, Validators.required],
            password: [null, Validators.required]
        });

        this.environment = environment;
        this.signaturesService.reset();
    }

    ionViewWillEnter() {
        this.menu.enable(false, 'left-menu');
        this.menu.enable(false, 'right-menu');
    }

    ngAfterViewInit(): void {
        setTimeout(async() => {
            this.showForm = true;
            this.fixAutoFill();
            await this.initConnection();
        }, 500);
    }

    fixAutoFill() {
        setTimeout(() => {
            this.loginForm.get('login').setValue($('#login').val());
            this.loginForm.get('password').setValue($('#password').val());
        }, 100);
    }

    async onSubmit(ssoToken = null) {
        const loading = await this.loadingController.create({
            cssClass: 'my-custom-class',
            message: this.translate.instant('lang.connexion'),
        });
        await loading.present();
        let url = '../rest/authenticate';
        if (ssoToken !== null) {
            url += ssoToken;
        }
        this.authService.authFailed = false;
        if (!this.loginForm.invalid ) {

            this.http.post(url, { 'login': this.loginForm.get('login').value, 'password': this.loginForm.get('password').value }, { observe: 'response' })
                .pipe(
                    tap((data: any) => {
                        this.loading = false;
                        this.showForm = false;
                        this.authService.saveTokens(data.headers.get('Token'), data.headers.get('Refresh-Token'));
                        this.authService.setUser({});

                        if (this.authService.getCachedUrl()) {
                            this.router.navigateByUrl(this.authService.getCachedUrl());
                            this.authService.cleanCachedUrl();
                        } else {
                            this.router.navigate(['/home']);
                        }
                        loading.dismiss();
                        this.hidePassword = true;
                    }),
                    catchError((err: any) => {
                        this.loading = false;
                        this.authService.authFailed = true;
                        if (err.status === 401 && ['kerberos', 'x509', 'azure_saml'].indexOf(this.authService.authMode) === -1)  {
                            this.notificationService.error('lang.wrongLoginPassword');
                        } else if (this.authService.authMode === 'azure_saml' && err.error.errors === 'Authentication Failed : not logged') {
                            window.location.href = err.error.authUri;
                        } else {
                            this.notificationService.handleErrors(err);
                        }
                        loading.dismiss();
                        return of(false);
                    })
                )
                .subscribe();
        } else {
            loading.dismiss();
            this.notificationService.error('lang.requiredLoginPassword');
        }
    }

    initConnection() {
        if (['kerberos', 'x509', 'azure_saml'].indexOf(this.authService.authMode) > -1) {
            this.loginForm.disable();
            this.loginForm.setValidators(null);
            this.onSubmit();
        } else if (['cas'].indexOf(this.authService.authMode) > -1) {
            this.loginForm.disable();
            this.loginForm.setValidators(null);
            if (this.authService.casToken != null) {
                const ssoToken = this.authService.casToken;
                this.authService.casToken = null;
                this.onSubmit(`?${ssoToken}`);
            } else {
                window.location.href = this.authService.authUri;
            }
        }
    }

    mailServerOffline() {
        this.notificationService.error(this.translate.instant('lang.mailServerOffline'));
    }
}
