import { Injectable, Injector } from '@angular/core';
import { HttpClient } from '@angular/common/http';
import { catchError, tap } from 'rxjs/operators';
import { Observable, of, Subject } from 'rxjs';
import { NotificationService } from '../notification.service';
import { TranslateService } from '@ngx-translate/core';
import { OtpFastService, OtpFieldsFast } from './otp-fast.service';
import { OtpFieldsYousign, OtpYousignService } from './otp-yousign.service';
import { FunctionsService } from '../functions.service';
import { Router } from '@angular/router';
import { AuthService } from '../auth.service';

@Injectable({
    providedIn: 'root'
})
export class OtpService {

    otpConnectorTypes: any[] = [
        {
            id: 'yousign',
            label: this.translate.instant('lang.yousignOtp'),
            logo: null
        },
        {
            id: 'fast',
            label: this.translate.instant('lang.fastOtp'),
            logo: null
        },
    ];

    status: any[] = [
        {
            id: 'SIGNED',
            message: this.translate.instant('lang.signedStatus'),
            hasAction: true
        },
        {
            id: 'REFUSED',
            message: this.translate.instant('lang.refusedStatus'),
            hasAction: true
        },
        {
            id: 'ERROR',
            message: this.translate.instant('lang.errorStatus'),
            hasAction: false
        },
        {
            id: 'NONE',
            message: this.translate.instant('lang.noneStatus'),
            hasAction: false
        }
    ];

    connectorService: OtpFastService | OtpYousignService;
    allowedConnectors: string[] = ['fast', 'yousign'];

    private eventAction = new Subject<any>();

    constructor(
        public http: HttpClient,
        public notificationService: NotificationService,
        public translate: TranslateService,
        private injector: Injector,
        private functions: FunctionsService,
        private router: Router,
        private authService: AuthService
    ) { }

    catchEvent(): Observable<any> {
        return this.eventAction.asObservable();
    }

    setEvent(content: any) {
        return this.eventAction.next(content);
    }

    setConnector(connectorType: any) {
        if (this.allowedConnectors.indexOf(connectorType?.id) > -1) {
            if (connectorType.id === 'fast') {
                this.connectorService = this.injector.get<OtpFastService>(OtpFastService);
            } else if (connectorType.id === 'yousign') {
                this.connectorService = this.injector.get<OtpYousignService>(OtpYousignService);
            }
        }
    }

    getUserOtpIcon(id: string): Promise<string> {
        if (!this.functions.empty(id)) {
            return new Promise((resolve) => {
                this.http.get(`assets/${id}.png`, { responseType: 'blob' }).pipe(
                    tap((response: any) => {
                        const reader = new FileReader();
                        reader.readAsDataURL(response);
                        reader.onloadend = () => {
                            resolve(reader.result as any);
                        };
                    }),
                    catchError(err => {
                        this.notificationService.handleErrors(err);
                        return of(false);
                    })
                ).subscribe();
            });
        }
    }

    getConnectorTypes(): any[] {
        return this.otpConnectorTypes;
    }

    getConnectorsConfig(type: string): Promise<any> {
        return new Promise((resolve) => {
            this.http.get('../rest/connectors').pipe(
                tap((data: any) => {
                    let result: any[] = data.otp;
                    if (!this.functions.empty(type)) {
                        result = result.filter((item: any) => item.type === type);
                    }
                    resolve(result);
                }),
                catchError((err: any) => {
                    this.notificationService.handleErrors(err);
                    resolve(null);
                    return of(false);
                })
            ).subscribe();
        });
    }

    getConnectorById(id: number): Promise<any> {
        return new Promise((resolve) => {
            this.http.get(`../rest/connectors/${id}`).pipe(
                tap((data: any) => {
                    resolve(data);
                }),
                catchError((err: any) => {
                    this.notificationService.handleErrors(err);
                    resolve(null);
                    return of(false);
                })
            ).subscribe();
        });
    }

    getOtpFields(): OtpFieldsFast | OtpFieldsYousign {
        return this.connectorService?.otpFields;
    }

    formatPhone() {
        if (this.connectorService?.otpFields?.phone?.length > 1 && this.connectorService?.otpFields?.phone[0] === '0') {
            this.connectorService.otpFields.phone = this.connectorService?.otpFields.phone.replace('0', '+33');
        }
    }

    getRegexPhone(): RegExp {
        // map country calling code with national number length
        const phonesMap = {
            '32': [8, 10],      // Belgium
            '41': [4, 12],      // Swiss
            '44': [7, 10],      // United Kingdom
            '352': [4, 11],     // Luxembourg
            '351': [9, 11],     // Portugal
            '33': 9,            // France
            '1' : 10,           // USA
            '39': 11,           // Italy
            '34': 9             // Spain
        };
        const regex = Object.keys(phonesMap).reduce((phoneFormats: any [], countryCode: any) => {
            const numberLength = phonesMap[countryCode];
            if (Array.isArray(numberLength)) {
                phoneFormats.push('(\\+' + countryCode + `[0-9]\{${numberLength[0]},${numberLength[1]}\})`);
            } else {
                phoneFormats.push('(\\+' + countryCode + `[0-9]\{${numberLength}\})`);
            }
            return phoneFormats;
        }, []).join('|');
        return new RegExp(`^(${regex})$`);
    }

    resetFieldsValues() {
        if (!this.functions.empty(this.connectorService?.otpFields)) {
            Object.keys(this.connectorService.otpFields).forEach((field: any) => {
                if (field !== 'modes' && field !== 'role' && field !== 'type') {
                    this.connectorService.otpFields[field] = null;
                }
            });
        }
    }

    getDocument(urlParams: {token: string; status: string}) {
        if (this.status.find((status: any) => status.id === urlParams.status && status.hasAction)) {
            this.http.get('../rest/externalSignatoryBookReturn', {params: {token: urlParams.token, status: urlParams.status}}).pipe(
                tap((data: any) => {
                    if (!this.functions.empty(data)) {
                        this.notificationService.message(this.status.find((status: any) => status.id === urlParams.status).message);
                    }
                }),
                catchError((err: any) => {
                    this.notificationService.handleErrors(err);
                    return of(false);
                })
            ).subscribe();
        } else {
            this.notificationService.error(this.status.find((status: any) => status.id === urlParams.status).message);
        }
        this.authService.clearTokens();
        this.router.navigate(['/login']);
    }
}
